package java.util.concurrent.atomic;

import java.util.function.LongBinaryOperator;
import java.util.function.LongUnaryOperator;
import sun.misc.Unsafe;
import java.io.Serializable;

public class AtomicLongArray implements Serializable
{
    private static final long serialVersionUID = -2308431214976778248L;
    private static final Unsafe unsafe;
    private static final int base;
    private static final int shift;
    private final long[] array;
    
    private long checkedByteOffset(int i) {
        if (i < 0 || i >= this.array.length) {
            throw new IndexOutOfBoundsException("index " + i);
        }
        return byteOffset(i);
    }
    
    private static long byteOffset(int n) {
        return ((long)n << AtomicLongArray.shift) + (long)AtomicLongArray.base;
    }
    
    public AtomicLongArray(int n) {
        this.array = new long[n];
    }
    
    public AtomicLongArray(long[] array) {
        this.array = (long[])array.clone();
    }
    
    public final int length() {
        return this.array.length;
    }
    
    public final long get(int n) {
        return this.getRaw(this.checkedByteOffset(n));
    }
    
    private long getRaw(long offset) {
        return AtomicLongArray.unsafe.getLongVolatile(this.array, offset);
    }
    
    public final void set(int n, long x) {
        AtomicLongArray.unsafe.putLongVolatile(this.array, this.checkedByteOffset(n), x);
    }
    
    public final void lazySet(int n, long x) {
        AtomicLongArray.unsafe.putOrderedLong(this.array, this.checkedByteOffset(n), x);
    }
    
    public final long getAndSet(int n, long newValue) {
        return AtomicLongArray.unsafe.getAndSetLong(this.array, this.checkedByteOffset(n), newValue);
    }
    
    public final boolean compareAndSet(int n, long n2, long n3) {
        return this.compareAndSetRaw(this.checkedByteOffset(n), n2, n3);
    }
    
    private boolean compareAndSetRaw(long offset, long expected, long x) {
        return AtomicLongArray.unsafe.compareAndSwapLong(this.array, offset, expected, x);
    }
    
    public final boolean weakCompareAndSet(int n, long n2, long n3) {
        return this.compareAndSet(n, n2, n3);
    }
    
    public final long getAndIncrement(int n) {
        return this.getAndAdd(n, 1L);
    }
    
    public final long getAndDecrement(int n) {
        return this.getAndAdd(n, -1L);
    }
    
    public final long getAndAdd(int n, long delta) {
        return AtomicLongArray.unsafe.getAndAddLong(this.array, this.checkedByteOffset(n), delta);
    }
    
    public final long incrementAndGet(int n) {
        return this.getAndAdd(n, 1L) + 1L;
    }
    
    public final long decrementAndGet(int n) {
        return this.getAndAdd(n, -1L) - 1L;
    }
    
    public long addAndGet(int n, long n2) {
        return this.getAndAdd(n, n2) + n2;
    }
    
    public final long getAndUpdate(int n, LongUnaryOperator longUnaryOperator) {
        long checkedByteOffset = this.checkedByteOffset(n);
        long raw;
        do {
            raw = this.getRaw(checkedByteOffset);
        } while (!this.compareAndSetRaw(checkedByteOffset, raw, longUnaryOperator.applyAsLong(raw)));
        return raw;
    }
    
    public final long updateAndGet(int n, LongUnaryOperator longUnaryOperator) {
        long checkedByteOffset = this.checkedByteOffset(n);
        long raw;
        long applyAsLong;
        do {
            raw = this.getRaw(checkedByteOffset);
            applyAsLong = longUnaryOperator.applyAsLong(raw);
        } while (!this.compareAndSetRaw(checkedByteOffset, raw, applyAsLong));
        return applyAsLong;
    }
    
    public final long getAndAccumulate(int n, long n2, LongBinaryOperator longBinaryOperator) {
        long checkedByteOffset = this.checkedByteOffset(n);
        long raw;
        do {
            raw = this.getRaw(checkedByteOffset);
        } while (!this.compareAndSetRaw(checkedByteOffset, raw, longBinaryOperator.applyAsLong(raw, n2)));
        return raw;
    }
    
    public final long accumulateAndGet(int n, long n2, LongBinaryOperator longBinaryOperator) {
        long checkedByteOffset = this.checkedByteOffset(n);
        long raw;
        long applyAsLong;
        do {
            raw = this.getRaw(checkedByteOffset);
            applyAsLong = longBinaryOperator.applyAsLong(raw, n2);
        } while (!this.compareAndSetRaw(checkedByteOffset, raw, applyAsLong));
        return applyAsLong;
    }
    
    @Override
    public String toString() {
        int n = this.array.length - 1;
        if (n == -1) {
            return "[]";
        }
        StringBuilder sb = new StringBuilder();
        sb.append('[');
        int n2 = 0;
        while (true) {
            sb.append(this.getRaw(byteOffset(n2)));
            if (n2 == n) {
                break;
            }
            sb.append(',').append(' ');
            ++n2;
        }
        return sb.append(']').toString();
    }
    
    static {
        unsafe = Unsafe.getUnsafe();
        base = AtomicLongArray.unsafe.arrayBaseOffset(long[].class);
        int arrayIndexScale = AtomicLongArray.unsafe.arrayIndexScale(long[].class);
        if ((arrayIndexScale & arrayIndexScale - 1) != 0x0) {
            throw new Error("data type scale not a power of two");
        }
        shift = 31 - Integer.numberOfLeadingZeros(arrayIndexScale);
        return;
    }
}
